@@ -33,9 +33,10 @@
 #include "src/missing.h"
 #include "src/eap.h"
 #include "src/eap-private.h"
+#include "src/iwd.h"
 
 static uint32_t default_mtu;
-struct l_queue *eap_methods;
+static struct l_queue *eap_methods;
 
 static void dump_eap(const char *str, void *user_data)
 {
@@ -110,14 +111,22 @@ bool eap_reset(struct eap_state *eap)
 	return true;
 }
 
-void eap_free(struct eap_state *eap)
+static void eap_free_common(struct eap_state *eap)
 {
 	if (eap->method_state && eap->method->free)
 		eap->method->free(eap);
 
-	if (eap->identity)
-		l_free(eap->identity);
+	eap->method = NULL;
 
+	if (eap->identity) {
+		l_free(eap->identity);
+		eap->identity = NULL;
+	}
+}
+
+void eap_free(struct eap_state *eap)
+{
+	eap_free_common(eap);
 	l_timeout_remove(eap->complete_timeout);
 
 	l_free(eap);
@@ -134,6 +143,11 @@ size_t eap_get_mtu(struct eap_state *eap)
 	return eap->mtu;
 }
 
+const char *eap_get_identity(struct eap_state *eap)
+{
+	return eap->identity;
+}
+
 /**
  * eap_send_response:
  * @eap: EAP state
@@ -546,13 +560,23 @@ bool eap_load_settings(struct eap_state *eap, struct l_settings *settings,
 		eap->identity = l_strdup(eap->method->get_identity(eap));
 	}
 
+	/*
+	 * RFC 4282 Section 2.2 - NAI Length Considerations
+	 *
+	 * Devices handling NAIs MUST support an NAI length of at least 72
+	 * octets. Support for an NAI length of 253 octets is RECOMMENDED.
+	 * ...
+	 * RADIUS is unable to support NAI lengths beyond 253 octets
+	 */
+	if (eap->identity && strlen(eap->identity) > 253) {
+		l_error("Identity is too long");
+		goto err;
+	}
+
 	return true;
 
 err:
-	if (eap->method_state && eap->method->free)
-		eap->method->free(eap);
-
-	eap->method = NULL;
+	eap_free_common(eap);
 
 	return false;
 }
@@ -580,13 +604,15 @@ const char *eap_get_method_name(struct eap_state *eap)
 void eap_set_key_material(struct eap_state *eap,
 				const uint8_t *msk_data, size_t msk_len,
 				const uint8_t *emsk_data, size_t emsk_len,
-				const uint8_t *iv, size_t iv_len)
+				const uint8_t *iv, size_t iv_len,
+				const uint8_t *session_id, size_t session_len)
 {
 	if (!eap->set_key_material)
 		return;
 
 	eap->set_key_material(msk_data, msk_len, emsk_data, emsk_len,
-				iv, iv_len, eap->user_data);
+				iv, iv_len, session_id, session_len,
+				eap->user_data);
 }
 
 void eap_method_event(struct eap_state *eap, unsigned int id, const void *data)
@@ -651,6 +677,12 @@ int eap_unregister_method(struct eap_method *method)
 	return -ENOENT;
 }
 
+void __eap_set_config(struct l_settings *config)
+{
+	if (!l_settings_get_uint(config, "EAP", "mtu", &default_mtu))
+		default_mtu = 1400; /* on WiFi the real MTU is around 2304 */
+}
+
 static void __eap_method_enable(struct eap_method_desc *start,
 					struct eap_method_desc *stop)
 {
@@ -690,7 +722,7 @@ static void __eap_method_disable(struct eap_method_desc *start,
 extern struct eap_method_desc __start___eap[];
 extern struct eap_method_desc __stop___eap[];
 
-void eap_init(uint32_t mtu)
+int eap_init(void)
 {
 	eap_methods = l_queue_new();
 	__eap_method_enable(__start___eap, __stop___eap);
@@ -700,10 +732,10 @@ void eap_init(uint32_t mtu)
 	 * EAP is capable of functioning on lower layers that
 	 *        provide an EAP MTU size of 1020 octets or greater.
 	 */
-	if (mtu == 0)
+	if (default_mtu == 0)
 		default_mtu = 1020;
-	else
-		default_mtu = mtu;
+
+	return 0;
 }
 
 void eap_exit(void)
@@ -711,3 +743,5 @@ void eap_exit(void)
 	__eap_method_disable(__start___eap, __stop___eap);
 	l_queue_destroy(eap_methods, NULL);
 }
+
+IWD_MODULE(eap, eap_init, eap_exit);
