@@ -159,15 +159,18 @@ static bool owe_compute_keys(struct owe_sm *owe, const void *public_key,
 
 	if (!l_ecdh_generate_shared_secret(owe->private, other_public,
 						&shared_secret)) {
+		l_ecc_point_free(other_public);
 		return false;
 	}
 
 	l_ecc_point_free(other_public);
 
 	nbytes = l_ecc_scalar_get_data(shared_secret, ss_buf, sizeof(ss_buf));
-
 	l_ecc_scalar_free(shared_secret);
 
+	if (nbytes < 0)
+		return false;
+
 	ptr += l_ecc_point_get_x(owe->public_key, ptr, sizeof(key));
 	memcpy(ptr, public_key, nbytes);
 	ptr += nbytes;
@@ -215,7 +218,6 @@ static bool owe_compute_keys(struct owe_sm *owe, const void *public_key,
 
 failed:
 	memset(ss_buf, 0, sizeof(ss_buf));
-	l_ecc_scalar_free(shared_secret);
 	return false;
 }
 
@@ -245,7 +247,7 @@ static int owe_rx_associate(struct auth_proto *ap, const uint8_t *frame,
 	size_t owe_dh_len = 0;
 	const uint8_t *owe_dh = NULL;
 	struct ie_rsn_info info;
-	bool akm_found;
+	bool akm_found = false;
 	const void *data;
 
 	mpdu = mpdu_validate(frame, len);
@@ -256,7 +258,8 @@ static int owe_rx_associate(struct auth_proto *ap, const uint8_t *frame,
 
 	body = mmpdu_body(mpdu);
 
-	if (body->status_code == MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP) {
+	if (L_LE16_TO_CPU(body->status_code) ==
+				MMPDU_STATUS_CODE_UNSUPP_FINITE_CYCLIC_GROUP) {
 		if (!owe_retry(owe))
 			goto owe_bad_status;
 
@@ -326,7 +329,7 @@ invalid_ies:
 	return MMPDU_STATUS_CODE_INVALID_ELEMENT;
 
 owe_bad_status:
-	return (int)body->status_code;
+	return L_LE16_TO_CPU(body->status_code);
 }
 
 struct auth_proto *owe_sm_new(struct handshake_state *hs,
