@@ -53,7 +53,6 @@
 #include "src/netconfig.h"
 #include "src/anqp.h"
 #include "src/anqputil.h"
-#include "src/hotspot.h"
 
 static struct l_queue *station_list;
 static uint32_t netdev_watch;
@@ -89,6 +88,8 @@ struct station {
 
 	struct l_queue *anqp_pending;
 
+	struct netconfig *netconfig;
+
 	bool preparing_roam : 1;
 	bool signal_low : 1;
 	bool roam_no_orig_ap : 1;
@@ -479,6 +480,8 @@ static void station_anqp_response_cb(enum anqp_result result,
 
 	known_networks_foreach(match_nai_realms, &search);
 
+	l_strv_free(realms);
+
 request_done:
 	l_queue_remove(station->anqp_pending, entry);
 
@@ -1157,6 +1160,7 @@ static void station_enter_state(struct station *station,
 	case STATION_STATE_DISCONNECTED:
 	case STATION_STATE_CONNECTED:
 		periodic_scan_stop(station);
+
 		break;
 	case STATION_STATE_DISCONNECTING:
 	case STATION_STATE_ROAMING:
@@ -1244,6 +1248,9 @@ static void station_disassociated(struct station *station)
 {
 	l_debug("%u", netdev_get_ifindex(station->netdev));
 
+	if (station->netconfig)
+		netconfig_reset(station->netconfig);
+
 	station_reset_connection_state(station);
 
 	station_enter_state(station, STATION_STATE_DISCONNECTED);
@@ -1278,6 +1285,11 @@ static void station_roamed(struct station *station)
 	station->signal_low = false;
 	station->roam_min_time.tv_sec = 0;
 	station->roam_no_orig_ap = false;
+
+	if (station->netconfig)
+		netconfig_reconfigure(station->netconfig);
+
+	station_enter_state(station, STATION_STATE_CONNECTED);
 }
 
 static void station_roam_failed(struct station *station)
@@ -1302,6 +1314,22 @@ static void station_roam_failed(struct station *station)
 		station_roam_timeout_rearm(station, 60);
 }
 
+static void station_netconfig_event_handler(enum netconfig_event event,
+							void *user_data)
+{
+	struct station *station = user_data;
+
+	switch (event) {
+	case NETCONFIG_EVENT_CONNECTED:
+		station_enter_state(station, STATION_STATE_CONNECTED);
+
+		break;
+	default:
+		l_error("station: Unsupported netconfig event: %d.", event);
+		break;
+	}
+}
+
 static void station_reassociate_cb(struct netdev *netdev,
 					enum netdev_result result,
 					void *event_data,
@@ -1314,10 +1342,9 @@ static void station_reassociate_cb(struct netdev *netdev,
 	if (station->state != STATION_STATE_ROAMING)
 		return;
 
-	if (result == NETDEV_RESULT_OK) {
+	if (result == NETDEV_RESULT_OK)
 		station_roamed(station);
-		station_enter_state(station, STATION_STATE_CONNECTED);
-	} else
+	else
 		station_roam_failed(station);
 }
 
@@ -1333,10 +1360,9 @@ static void station_fast_transition_cb(struct netdev *netdev,
 	if (station->state != STATION_STATE_ROAMING)
 		return;
 
-	if (result == NETDEV_RESULT_OK) {
+	if (result == NETDEV_RESULT_OK)
 		station_roamed(station);
-		station_enter_state(station, STATION_STATE_CONNECTED);
-	} else
+	else
 		station_roam_failed(station);
 }
 
@@ -1695,6 +1721,25 @@ static void station_roam_scan(struct station *station,
 		station_roam_failed(station);
 }
 
+static bool station_roam_scan_known_freqs(struct station *station)
+{
+	const struct network_info *info = network_get_info(
+						station->connected_network);
+	struct scan_freq_set *freqs = network_info_get_roam_frequencies(info,
+					station->connected_bss->frequency, 5);
+
+	if (!freqs) {
+		l_debug("no known frequencies to scan");
+		return false;
+	}
+
+	station_roam_scan(station, freqs);
+
+	scan_freq_set_free(freqs);
+
+	return true;
+}
+
 static uint32_t station_freq_from_neighbor_report(const uint8_t *country,
 		struct ie_neighbor_report_info *info, enum scan_band *out_band)
 {
@@ -1764,8 +1809,10 @@ static void station_neighbor_report_cb(struct netdev *netdev, int err,
 		return;
 
 	if (!reports || err) {
-		/* Have to do a full scan */
-		station_roam_scan(station, NULL);
+		if (!station_roam_scan_known_freqs(station)) {
+			l_debug("no neighbor report results or known freqs");
+			station_roam_failed(station);
+		}
 
 		return;
 	}
@@ -1894,8 +1941,11 @@ static void station_roam_trigger_cb(struct l_timeout *timeout, void *user_data)
 						station_neighbor_report_cb))
 			return;
 
-	/* Otherwise do a full scan for target BSS candidates */
-	station_roam_scan(station, NULL);
+	if (!station_roam_scan_known_freqs(station)) {
+		l_debug("No neighbor report or known frequencies, roam failed");
+		station_roam_failed(station);
+		return;
+	}
 }
 
 static void station_roam_timeout_rearm(struct station *station, int seconds)
@@ -2232,18 +2282,23 @@ static void station_connect_cb(struct netdev *netdev, enum netdev_result result,
 	}
 
 	network_connected(station->connected_network);
-	station_enter_state(station, STATION_STATE_CONNECTED);
+
+	if (station->netconfig)
+		netconfig_configure(station->netconfig,
+					network_get_settings(
+						station->connected_network),
+					netdev_get_address(station->netdev),
+					station_netconfig_event_handler,
+					station);
+	else
+		station_enter_state(station, STATION_STATE_CONNECTED);
 }
 
 int __station_connect_network(struct station *station, struct network *network,
 				struct scan_bss *bss)
 {
-	const uint8_t *rc = NULL;
-	uint8_t hs20_ie[7];
-	size_t rc_len = 0;
-	uint8_t rc_buf[32];
-	struct iovec iov[2];
-	int iov_elems = 0;
+	const struct iovec *extra_ies;
+	size_t iov_elems = 0;
 	struct handshake_state *hs;
 	int r;
 
@@ -2251,30 +2306,9 @@ int __station_connect_network(struct station *station, struct network *network,
 	if (!hs)
 		return -ENOTSUP;
 
-	if (bss->hs20_capable) {
-		/* Include HS20 Indication with (Re)Association */
-		ie_build_hs20_indication(bss->hs20_version, hs20_ie);
+	extra_ies = network_get_extra_ies(network, &iov_elems);
 
-		iov[iov_elems].iov_base = hs20_ie;
-		iov[iov_elems].iov_len = hs20_ie[1] + 2;
-		iov_elems++;
-
-		/*
-		 * If a matching roaming consortium OI is found for the network
-		 * this single RC value will be set in the handshake and used
-		 * during (Re)Association.
-		 */
-		rc = hs20_get_roaming_consortium(network, &rc_len);
-		if (rc) {
-			ie_build_roaming_consortium(rc, rc_len, rc_buf);
-
-			iov[iov_elems].iov_base = rc_buf;
-			iov[iov_elems].iov_len = rc_buf[1] + 2;
-			iov_elems++;
-		}
-	}
-
-	r = netdev_connect(station->netdev, bss, hs, iov_elems ? iov : NULL,
+	r = netdev_connect(station->netdev, bss, hs, extra_ies,
 				iov_elems, station_netdev_event,
 				station_connect_cb, station);
 	if (r < 0) {
@@ -2325,6 +2359,9 @@ static void station_disconnect_onconnect(struct station *station,
 		return;
 	}
 
+	if (station->netconfig)
+		netconfig_reset(station->netconfig);
+
 	station_reset_connection_state(station);
 
 	station_enter_state(station, STATION_STATE_DISCONNECTING);
@@ -2560,6 +2597,9 @@ int station_disconnect(struct station *station)
 					station_disconnect_cb, station) < 0)
 		return -EIO;
 
+	if (station->netconfig)
+		netconfig_reset(station->netconfig);
+
 	/*
 	 * If the disconnect somehow fails we won't know if we're still
 	 * connected so we may as well indicate now that we're no longer
@@ -3021,7 +3061,7 @@ static struct station *station_create(struct netdev *netdev)
 	l_dbus_object_add_interface(dbus, netdev_get_path(netdev),
 					IWD_STATION_INTERFACE, station);
 
-	netconfig_ifindex_add(netdev_get_ifindex(netdev));
+	station->netconfig = netconfig_new(netdev_get_ifindex(netdev));
 
 	station->anqp_pending = l_queue_new();
 
@@ -3038,7 +3078,10 @@ static void station_free(struct station *station)
 	if (station->connected_bss)
 		netdev_disconnect(station->netdev, NULL, NULL);
 
-	netconfig_ifindex_remove(netdev_get_ifindex(station->netdev));
+	if (station->netconfig) {
+		netconfig_destroy(station->netconfig);
+		station->netconfig = NULL;
+	}
 
 	periodic_scan_stop(station);
 
@@ -3166,3 +3209,4 @@ static void station_exit(void)
 }
 
 IWD_MODULE(station, station_init, station_exit)
+IWD_MODULE_DEPENDS(station, netconfig)
